@@ -14,15 +14,17 @@
 import concurrent.futures
 import inspect
 import json
+import multiprocessing as mp
 import os
 import shutil
 import tempfile
+from collections.abc import Sequence
 from dataclasses import dataclass
 from datetime import datetime
 from functools import partial
 from pathlib import Path
 from types import FunctionType
-from typing import TYPE_CHECKING, Any, Callable, Dict, List, Literal, Optional, Sequence, Tuple, Union
+from typing import TYPE_CHECKING, Any, Callable, Literal, Optional, Union
 from urllib import parse
 
 import torch
@@ -61,7 +63,7 @@ def _is_remote_file(path: str) -> bool:
     return obj.scheme in _SUPPORTED_PROVIDERS
 
 
-def _get_indexed_paths(data: Any) -> Dict[int, str]:
+def _get_indexed_paths(data: Any) -> dict[int, str]:
     flattened_item, _ = tree_flatten(data)
 
     return {
@@ -110,7 +112,7 @@ class LambdaMapRecipe(MapRecipe):
         self,
         fn: Callable[[str, Any], None],
         inputs: Union[Sequence[Any], StreamingDataLoader],
-        storage_options: Dict[str, Any] = {},
+        storage_options: dict[str, Any] = {},
     ):
         super().__init__(storage_options)
         self._fn = fn
@@ -129,7 +131,7 @@ class LambdaMapRecipe(MapRecipe):
         if self._contains_device and self._device is None:
             self._find_device()
 
-        kwargs: Dict[str, Any] = {}
+        kwargs: dict[str, Any] = {}
 
         if self._contains_device:
             kwargs["device"] = self._device
@@ -164,8 +166,8 @@ class LambdaDataChunkRecipe(DataChunkRecipe):
         chunk_bytes: Optional[Union[int, str]],
         compression: Optional[str],
         encryption: Optional[Encryption] = None,
-        existing_index: Optional[Dict[str, Any]] = None,
-        storage_options: Dict[str, Any] = {},
+        existing_index: Optional[dict[str, Any]] = None,
+        storage_options: dict[str, Any] = {},
     ):
         super().__init__(
             chunk_size=chunk_size,
@@ -204,12 +206,45 @@ class LambdaDataChunkRecipe(DataChunkRecipe):
         """Being overridden dynamically."""
 
 
+class QueueDataChunkRecipe(DataChunkRecipe):
+    """A recipe to optimize data using a shared `multiprocessing.Queue`."""
+
+    def __init__(
+        self,
+        fn: Callable[[Any], None],
+        queue: mp.Queue,
+        chunk_size: Optional[int],
+        chunk_bytes: Optional[Union[int, str]],
+        compression: Optional[str],
+        encryption: Optional[Encryption] = None,
+        existing_index: Optional[dict[str, Any]] = None,
+        storage_options: dict[str, Any] = {},
+    ):
+        super().__init__(
+            chunk_size=chunk_size,
+            chunk_bytes=chunk_bytes,
+            compression=compression,
+            encryption=encryption,
+            storage_options=storage_options,
+        )
+        self._fn = fn
+        self._queue = queue
+        self.existing_index = existing_index
+        self.is_generator = False
+
+    def prepare_structure(self, input_dir: Optional[str]) -> Any:
+        return self._queue
+
+    def prepare_item(self, item_metadata: Any) -> Any:
+        return self._fn(item_metadata)
+
+
 def map(
     fn: Callable[[str, Any], None],
     inputs: Union[Sequence[Any], StreamingDataLoader],
-    output_dir: Union[str, Dir],
-    input_dir: Optional[str] = None,
-    weights: Optional[List[int]] = None,
+    output_dir: Union[str, Path, Dir],
+    input_dir: Optional[Union[str, Path]] = None,
+    weights: Optional[list[int]] = None,
     num_workers: Optional[int] = None,
     fast_dev_run: Union[bool, int] = False,
     num_nodes: Optional[int] = None,
@@ -222,7 +257,8 @@ def map(
     batch_size: Optional[int] = None,
     start_method: Optional[str] = None,
     optimize_dns: Optional[bool] = None,
-    storage_options: Dict[str, Any] = {},
+    storage_options: dict[str, Any] = {},
+    keep_data_ordered: bool = True,
 ) -> None:
     """Maps a callable over a collection of inputs, possibly in a distributed way.
 
@@ -248,6 +284,11 @@ def map(
             inside an interactive shell like Ipython.
         optimize_dns: Whether the optimized dns should be used.
         storage_options: Storage options for the cloud provider.
+        keep_data_ordered (bool): Whether to use a shared queue for item distribution among workers.
+            If False, all workers will fetch items dynamically from a shared queue, which helps balance
+            workload and reduce idle time when some workers finish early. This may lead to unordered
+            processing of items. If True, each worker processes a statically assigned subset of items
+            in order.
     """
     _check_version_and_prompt_upgrade(__version__)
 
@@ -313,6 +354,7 @@ def map(
             reader=reader,
             start_method=start_method,
             storage_options=storage_options,
+            keep_data_ordered=keep_data_ordered,
         )
 
         with optimize_dns_context(optimize_dns if optimize_dns is not None else False):
@@ -344,10 +386,11 @@ def map(
 #
 def optimize(
     fn: Callable[[Any], Any],
-    inputs: Union[Sequence[Any], StreamingDataLoader],
-    output_dir: str,
+    inputs: Optional[Union[Sequence[Any], StreamingDataLoader]] = None,
+    output_dir: str = "optimized_data",
+    queue: Optional[mp.Queue] = None,
     input_dir: Optional[str] = None,
-    weights: Optional[List[int]] = None,
+    weights: Optional[list[int]] = None,
     chunk_size: Optional[int] = None,
     chunk_bytes: Optional[Union[int, str]] = None,
     compression: Optional[str] = None,
@@ -366,7 +409,9 @@ def optimize(
     item_loader: Optional[BaseItemLoader] = None,
     start_method: Optional[str] = None,
     optimize_dns: Optional[bool] = None,
-    storage_options: Dict[str, Any] = {},
+    storage_options: dict[str, Any] = {},
+    keep_data_ordered: bool = True,
+    verbose: bool = True,
 ) -> None:
     """This function converts a dataset into chunks, possibly in a distributed way.
 
@@ -375,6 +420,8 @@ def optimize(
             corresponds to the input. Every invocation of the function should return a similar hierarchy of objects,
             where the object types and list sizes don't change.
         inputs: A sequence of input to be processed by the `fn` function, or a streaming dataloader.
+        queue: A multiprocessing queue to be used for processing the inputs. If provided, the `inputs` argument
+            will be ignored, and the function will process the items from the queue instead.
         output_dir: The folder where the processed data should be stored.
         input_dir: Provide the path where your files are stored. If the files are on a remote storage,
             they will be downloaded in the background while processed.
@@ -403,9 +450,24 @@ def optimize(
             inside an interactive shell like Ipython.
         optimize_dns: Whether the optimized dns should be used.
         storage_options: Storage options for the cloud provider.
+        keep_data_ordered (bool): Whether to use a shared queue for item distribution among workers.
+            If False, all workers will fetch items dynamically from a shared queue, which helps balance
+            workload and reduce idle time when some workers finish early. This may lead to unordered
+            processing of items. If True, each worker processes a statically assigned subset of items
+            in order.
+        verbose: Whether to print the progress of the optimization. Defaults to True.
     """
     _check_version_and_prompt_upgrade(__version__)
 
+    if queue is not None and inputs is not None:
+        raise ValueError("You can either provide `inputs` or `queue`, not both.")
+
+    if queue is None and inputs is None:
+        raise ValueError("At least one of `inputs` or `queue` must be provided.")
+
+    if queue is not None and not isinstance(queue, mp.queues.Queue):
+        raise ValueError(f"The provided `queue` should be a multiprocessing queue. Found {type(queue)}.")
+
     if mode is not None and mode not in ["append", "overwrite"]:
         raise ValueError(f"The provided `mode` should be either `append` or `overwrite`. Found {mode}.")
 
@@ -415,13 +477,15 @@ def optimize(
     if isinstance(inputs, StreamingDataLoader) and weights is not None:
         raise ValueError("When providing a streaming dataloader, weights isn't supported.")
 
-    if not isinstance(inputs, (Sequence, StreamingDataLoader)):
-        raise ValueError(
-            f"The provided inputs should be a non-empty sequence or a streaming dataloader. Found {inputs}."
-        )
+    if queue is None:
+        if not isinstance(inputs, (Sequence, StreamingDataLoader)):
+            raise ValueError(
+                f"The provided inputs should be a non-empty sequence or a streaming dataloader. Found {inputs}."
+            )
+
+        if len(inputs) == 0:
+            raise ValueError(f"The provided inputs should be non empty. Found {inputs}.")
 
-    if len(inputs) == 0:
-        raise ValueError(f"The provided inputs should be non empty. Found {inputs}.")
     if chunk_size is None and chunk_bytes is None:
         raise ValueError("Either `chunk_size` or `chunk_bytes` needs to be defined.")
 
@@ -430,7 +494,7 @@ def optimize(
             "Only https://lightning.ai/ supports multiple nodes or selecting a machine.Create an account to try it out."
         )
 
-    if not _IS_IN_STUDIO:
+    if not _IS_IN_STUDIO and verbose:
         print(
             "Create an account on https://lightning.ai/ to optimize your data faster "
             "using multiple nodes and large machines."
@@ -461,7 +525,8 @@ def optimize(
             _output_dir, mode=mode, use_checkpoint=use_checkpoint, storage_options=storage_options
         )
 
-        if not isinstance(inputs, StreamingDataLoader):
+        if not isinstance(inputs, StreamingDataLoader) and queue is None:
+            assert inputs is not None
             resolved_dir = _resolve_dir(input_dir or _get_input_dir(inputs))
 
             if isinstance(batch_size, int) and batch_size > 1:
@@ -473,7 +538,7 @@ def optimize(
             num_workers = 1
 
         num_workers = num_workers or _get_default_num_workers()
-        state_dict = {rank: 0 for rank in range(num_workers)}
+        state_dict = dict.fromkeys(range(num_workers), 0)
 
         existing_index_file_content = (
             read_index_file_content(_output_dir, storage_options) if mode == "append" else None
@@ -500,11 +565,15 @@ def optimize(
             item_loader=item_loader,
             start_method=start_method,
             storage_options=storage_options,
+            keep_data_ordered=keep_data_ordered,
+            verbose=verbose,
         )
 
         with optimize_dns_context(optimize_dns if optimize_dns is not None else False):
-            data_processor.run(
-                LambdaDataChunkRecipe(
+            recipe: Optional[Union[LambdaDataChunkRecipe, QueueDataChunkRecipe]] = None
+            if queue is None:
+                assert isinstance(inputs, (Sequence, StreamingDataLoader))
+                recipe = LambdaDataChunkRecipe(
                     fn,
                     inputs,
                     chunk_size=chunk_size,
@@ -514,7 +583,20 @@ def optimize(
                     existing_index=existing_index_file_content,
                     storage_options=storage_options,
                 )
-            )
+            else:
+                assert queue is not None
+                recipe = QueueDataChunkRecipe(
+                    fn,
+                    queue,
+                    chunk_size=chunk_size,
+                    chunk_bytes=chunk_bytes,
+                    compression=compression,
+                    encryption=encryption,
+                    existing_index=existing_index_file_content,
+                    storage_options=storage_options,
+                )
+            assert recipe is not None, "Recipe should be defined at this point."
+            data_processor.run(recipe)
         return None
     return _execute(
         f"litdata-optimize-{datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}",
@@ -523,7 +605,7 @@ def optimize(
     )
 
 
-def _listdir(folder: str) -> Tuple[str, List[str]]:
+def _listdir(folder: str) -> tuple[str, list[str]]:
     return folder, os.listdir(folder)
 
 
@@ -537,7 +619,7 @@ class walk:
     def __init__(self, folder: str, max_workers: Optional[int] = os.cpu_count()) -> None:
         self.folders = [folder]
         self.max_workers = max_workers or 1
-        self.futures: List[concurrent.futures.Future] = []
+        self.futures: list[concurrent.futures.Future] = []
 
         if not _IS_IN_STUDIO:
             print("This method is optimized to run on https://lightning.ai/. Don't use it otherwise.")
@@ -582,10 +664,10 @@ class CopyInfo:
 
 
 def merge_datasets(
-    input_dirs: List[str],
+    input_dirs: list[str],
     output_dir: str,
     max_workers: Optional[int] = os.cpu_count(),
-    storage_options: Dict[str, Any] = {},
+    storage_options: dict[str, Any] = {},
 ) -> None:
     """Enables to merge multiple existing optimized datasets into a single optimized dataset.
 
@@ -628,7 +710,7 @@ def merge_datasets(
             raise ValueError("Your are trying to merge datasets with different compression configuration.")
 
     chunks = []
-    copy_infos: List[CopyInfo] = []
+    copy_infos: list[CopyInfo] = []
     counter = 0
     for input_dir, input_dir_file_content in zip(resolved_input_dirs, input_dirs_file_content):
         compression = input_dir_file_content["config"]["compression"]  # type: ignore
@@ -648,7 +730,7 @@ def merge_datasets(
     _tqdm = _get_tqdm_iterator_if_available()
 
     with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
-        futures: List[concurrent.futures.Future] = []
+        futures: list[concurrent.futures.Future] = []
         for copy_info in copy_infos:
             future = executor.submit(_apply_copy, copy_info, resolved_output_dir, storage_options)
             futures.append(future)
@@ -659,7 +741,7 @@ def merge_datasets(
     _save_index(index_json, resolved_output_dir, storage_options)
 
 
-def _apply_copy(copy_info: CopyInfo, output_dir: Dir, storage_options: Dict[str, Any] = {}) -> None:
+def _apply_copy(copy_info: CopyInfo, output_dir: Dir, storage_options: dict[str, Any] = {}) -> None:
     if output_dir.url is None and copy_info.input_dir.url is None:
         assert copy_info.input_dir.path
         assert output_dir.path
@@ -678,7 +760,7 @@ def _apply_copy(copy_info: CopyInfo, output_dir: Dir, storage_options: Dict[str,
         raise NotImplementedError
 
 
-def _save_index(index_json: Dict, output_dir: Dir, storage_options: Dict[str, Any] = {}) -> None:
+def _save_index(index_json: dict, output_dir: Dir, storage_options: dict[str, Any] = {}) -> None:
     if output_dir.url is None:
         assert output_dir.path
         with open(os.path.join(output_dir.path, _INDEX_FILENAME), "w") as f:
