@@ -14,7 +14,7 @@
 import contextlib
 import os
 from abc import ABC, abstractmethod
-from typing import Any, List
+from typing import Any
 
 from litdata.constants import _PYARROW_AVAILABLE
 from litdata.streaming.dataloader import StreamingDataLoader
@@ -49,7 +49,7 @@ class BaseReader(ABC):
         return int(os.getenv("DATA_OPTIMIZER_NODE_RANK", 0))
 
     @abstractmethod
-    def remap_items(self, items: Any, num_workers: int) -> List[Any]:
+    def remap_items(self, items: Any, num_workers: int) -> list[Any]:
         """Remap the items provided by the users into items more adapted to be distributed."""
 
     @abstractmethod
@@ -92,7 +92,7 @@ class ParquetReader(BaseReader):
         self.parquet_file = pq.ParquetFile(filepath, memory_map=True)
         return self.parquet_file
 
-    def remap_items(self, filepaths: List[str], _: int) -> List[str]:
+    def remap_items(self, filepaths: list[str], _: int) -> list[str]:
         """Reshard the parquet files for optimized processing.
 
         If a parquet file contains more number of rows than a specified `num_rows`,
@@ -143,10 +143,25 @@ class StreamingDataLoaderReader(BaseReader):
         """Read the next item from the dataloader."""
         if self.dataloader_iter is None:
             self.dataloader_iter = iter(self.dataloader)
-        return next(self.dataloader_iter)
 
-    def remap_items(self, dataloader: StreamingDataLoader, _: int) -> List[Any]:
-        """Remap the items from the dataloader. But here, we don't do anything.
-        No splitting or batching is done here. As streaming dataloader is already optimized for this.
+        try:
+            # Data is distributed across workers through iterator, similar to DDP.
+            # Although the iterator is created within this worker process,
+            # distribution is already managed by the StreamingDataLoader and StreamingDataset.
+            return next(self.dataloader_iter)
+        except StopIteration:
+            # This can happen when some workers finish their data slice before others.
+            # In multiprocessing scenarios with StreamingDataLoader, this is expected behavior.
+            # We return None to signal that this worker has no more data to process.
+            return None
+
+    def remap_items(self, items: Any, num_workers: int) -> list[Any]:
+        """For StreamingDataLoader, we need to be smarter about item distribution.
+        We create enough virtual items so that each worker can process until its
+        portion of the dataloader is exhausted.
         """
-        return list(range(len(dataloader)))
+        # The items parameter is the StreamingDataLoader in this case
+        total_items = len(items)
+        # Create more virtual items than actual items to ensure all workers
+        # get enough chances to process their portion of the dataloader
+        return list(range(total_items * 2))
