@@ -11,13 +11,17 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+import logging
 from abc import ABC, abstractmethod
-from typing import Dict, TypeVar
+from typing import TypeVar
 
 from litdata.constants import _ZSTD_AVAILABLE
+from litdata.debugger import ChromeTraceColors, _get_log_msg
 
 TCompressor = TypeVar("TCompressor", bound="Compressor")
 
+logger = logging.getLogger("litdata.streaming.compression")
+
 
 class Compressor(ABC):
     """Base class for compression algorithm."""
@@ -32,7 +36,7 @@ class Compressor(ABC):
 
     @classmethod
     @abstractmethod
-    def register(cls, compressors: Dict[str, "Compressor"]) -> None:
+    def register(cls, compressors: dict[str, "Compressor"]) -> None:
         pass
 
 
@@ -58,10 +62,13 @@ class ZSTDCompressor(Compressor):
     def decompress(self, data: bytes) -> bytes:
         import zstd
 
-        return zstd.decompress(data)
+        logger.debug(_get_log_msg({"name": "Decompressing data", "ph": "B", "cname": ChromeTraceColors.MUSTARD_YELLOW}))
+        decompressed_data = zstd.decompress(data)
+        logger.debug(_get_log_msg({"name": "Decompressed data", "ph": "E", "cname": ChromeTraceColors.MUSTARD_YELLOW}))
+        return decompressed_data
 
     @classmethod
-    def register(cls, compressors: Dict[str, "Compressor"]) -> None:
+    def register(cls, compressors: dict[str, "Compressor"]) -> None:
         if not _ZSTD_AVAILABLE:
             return
 
@@ -72,6 +79,6 @@ class ZSTDCompressor(Compressor):
             compressors[f"zstd:{level}"] = ZSTDCompressor(level)
 
 
-_COMPRESSORS: Dict[str, Compressor] = {}
+_COMPRESSORS: dict[str, Compressor] = {}
 
 ZSTDCompressor.register(_COMPRESSORS)
