@@ -14,13 +14,14 @@
 import logging
 import os
 from time import time
-from typing import Any, Dict, List, Optional, Tuple, Union
+from typing import Any, Callable, Optional, Union
 
 import numpy as np
 from torch.utils.data import IterableDataset
 
 from litdata import __version__
 from litdata.constants import _INDEX_FILENAME
+from litdata.debugger import _get_log_msg
 from litdata.helpers import _check_version_and_prompt_upgrade
 from litdata.streaming import Cache
 from litdata.streaming.item_loader import BaseItemLoader, ParquetLoader
@@ -38,7 +39,7 @@ from litdata.utilities.shuffle import (
     _map_node_worker_rank_to_chunk_indexes_to_not_delete,
 )
 
-logger = logging.getLogger(__name__)
+logger = logging.getLogger("litdata.streaming.dataset")
 
 
 class StreamingDataset(IterableDataset):
@@ -52,19 +53,22 @@ class StreamingDataset(IterableDataset):
         shuffle: bool = False,
         drop_last: Optional[bool] = None,
         seed: int = 42,
-        serializers: Optional[Dict[str, Serializer]] = None,
+        serializers: Optional[dict[str, Serializer]] = None,
         max_cache_size: Union[int, str] = "100GB",
         subsample: float = 1.0,
         encryption: Optional[Encryption] = None,
-        storage_options: Optional[Dict] = {},
+        storage_options: Optional[dict] = {},
+        session_options: Optional[dict] = {},
         max_pre_download: int = 2,
         index_path: Optional[str] = None,
         force_override_state_dict: bool = False,
+        transform: Optional[Union[Callable, list[Callable]]] = None,
     ) -> None:
         """The streaming dataset can be used once your data have been optimised using the DatasetOptimiser class.
 
         Args:
-            input_dir: Path to the folder where the input data is stored.
+            input_dir: Path to the folder where the input data is stored. Supports paths ending with `.parquet`
+                with wildcards in the basename to stream specific Parquet files.
             cache_dir: Path to the folder where the cache data is stored. If not provided, the cache will be stored
                 in the default cache directory.
             item_loader: The logic to load an item from a chunk.
@@ -79,12 +83,13 @@ class StreamingDataset(IterableDataset):
             subsample: Float representing fraction of the dataset to be randomly sampled (e.g., 0.1 => 10% of dataset).
             encryption: The encryption object to use for decrypting the data.
             storage_options: Additional connection options for accessing storage services.
+            session_options: Additional connection options for accessing S3 services.
             max_pre_download: Maximum number of chunks that can be pre-downloaded by the StreamingDataset.
             index_path: Path to `index.json` for the Parquet dataset.
                 If `index_path` is a directory, the function will look for `index.json` within it.
                 If `index_path` is a full file path, it will use that directly.
             force_override_state_dict: Boolean flag for allowing local arguments to override a loaded state dict.
-
+            transform: Optional transformation function or list of functions to apply to each item in the dataset.
         """
         _check_version_and_prompt_upgrade(__version__)
 
@@ -95,16 +100,17 @@ class StreamingDataset(IterableDataset):
         if not isinstance(subsample, float) or subsample <= 0.0:
             raise ValueError("subsample must be a float with value greater than 0.")
 
+        fnmatch_pattern = None
+        if isinstance(input_dir, str) and input_dir.endswith(".parquet"):
+            input_dir, fnmatch_pattern = os.path.split(input_dir)
+
         input_dir = _resolve_dir(input_dir)
         cache_dir = _resolve_dir(cache_dir)
 
         if input_dir.url is not None and input_dir.url.startswith("hf://"):
             if index_path is None:
                 # No index_path was provided. Attempt to load it from cache or generate it dynamically on the fly.
-                index_path = index_hf_dataset(input_dir.url)
-                cache_dir.path = index_path
-                input_dir.path = index_path
-
+                index_path = index_hf_dataset(input_dir.url, cache_dir.path)
             if item_loader is not None and not isinstance(item_loader, ParquetLoader):
                 raise ValueError(
                     "Invalid item_loader for hf://datasets. "
@@ -116,10 +122,19 @@ class StreamingDataset(IterableDataset):
 
         self.input_dir = input_dir
         self.cache_dir = cache_dir
-        self.subsampled_files: List[str] = []
-        self.region_of_interest: List[Tuple[int, int]] = []
+        self.subsampled_files: list[str] = []
+        self.region_of_interest: list[tuple[int, int]] = []
         self.subsampled_files, self.region_of_interest = subsample_streaming_dataset(
-            self.input_dir, self.cache_dir, item_loader, subsample, shuffle, seed, storage_options, index_path
+            self.input_dir,
+            self.cache_dir,
+            item_loader,
+            subsample,
+            shuffle,
+            seed,
+            storage_options,
+            session_options,
+            index_path,
+            fnmatch_pattern,
         )
 
         self.item_loader = item_loader
@@ -154,28 +169,59 @@ class StreamingDataset(IterableDataset):
 
         self.cache: Optional[Cache] = None
         self.worker_env: Optional[_WorkerEnv] = None
-        self.worker_chunks: List[int] = []
-        self.worker_intervals: List[List[int]] = []
-        self.current_indexes: List[int] = []
-        self.chunk_index = 0
-        self.num_chunks: Optional[int] = None
-        self.global_index = 0
-        self.index = 0
+        self.worker_chunks: list[int] = []  # chunk indexes that the current worker will download, read & stream
+        self.worker_intervals: list[list[int]] = []  # chunk index intervals for the current worker
+        self.upcoming_indexes: list[int] = []  # contains list of upcoming indexes to be processed
+
+        # which index of the array `self.worker_chunks` will we work on after this chunk is completely consumed
+        self.worker_next_chunk_index = 0
+
+        self.num_chunks: Optional[int] = None  # total number of chunks that the current worker will work on
+        self.global_index = 0  # total number of samples processed by the current worker up until now
+
+        # number of samples processed by the current worker in the current chunk
+        self.consumed_sample_count_in_curr_chunk = 0
         self.has_triggered_download = False
         self.min_items_per_replica: Optional[int] = None
         self.current_epoch = 1
         self.random_state = None
         self.shuffler: Optional[Shuffle] = None
         self.serializers = serializers
-        self._state_dict: Optional[Dict[str, Any]] = None
+        self._state_dict: Optional[dict[str, Any]] = None
         self._force_override_state_dict = force_override_state_dict
         # Has slightly different meaning in the context of the dataset
         # We consider `num_workers = 0` from `torch.utils.DataLoader` still as 1 worker (the main process)
         self.num_workers: int = 1
         self.batch_size: int = 1
         self._encryption = encryption
+        # Ensure data_connection_id is included in storage_options if available from input_dir
+        if input_dir.data_connection_id and storage_options is not None:
+            storage_options = storage_options.copy()
+            storage_options["data_connection_id"] = input_dir.data_connection_id
+        elif input_dir.data_connection_id and storage_options is None:
+            storage_options = {"data_connection_id": input_dir.data_connection_id}
         self.storage_options = storage_options
+        self.session_options = session_options
         self.max_pre_download = max_pre_download
+        if transform is not None:
+            transform = transform if isinstance(transform, list) else [transform]
+            for t in transform:
+                if not callable(t):
+                    raise ValueError(f"Transform should be a callable. Found {t}")
+            self.transform = transform
+        self._on_demand_bytes = True  # true by default, when iterating, turn this off to store the chunks in the cache
+
+    @property
+    def on_demand_bytes(self) -> bool:
+        return self._on_demand_bytes
+
+    @on_demand_bytes.setter
+    def on_demand_bytes(self, value: bool) -> None:
+        if not isinstance(value, bool):
+            raise ValueError(f"on_demand_bytes should be a boolean. Found {value}")
+        self._on_demand_bytes = value
+        assert self.cache is not None, "Cache must be initialized before setting on_demand_bytes."
+        self.cache._reader.on_demand_bytes = value
 
     def set_shuffle(self, shuffle: bool) -> None:
         self.shuffle = shuffle
@@ -213,7 +259,9 @@ class StreamingDataset(IterableDataset):
             max_cache_size=self.max_cache_size,
             encryption=self._encryption,
             storage_options=self.storage_options,
+            session_options=self.session_options,
             max_pre_download=self.max_pre_download,
+            on_demand_bytes=self._on_demand_bytes,
         )
         cache._reader._try_load_config()
 
@@ -229,7 +277,7 @@ class StreamingDataset(IterableDataset):
         seed = self.seed
         drop_last = self.drop_last
         if self._state_dict is not None:
-            state: Dict[str, Any] = self._state_dict
+            state: dict[str, Any] = self._state_dict
             seed = state["seed"]
             drop_last = state["drop_last"]
         return FullShuffle(cache, seed, drop_last) if self.shuffle else NoShuffle(cache, seed, drop_last)
@@ -254,17 +302,19 @@ class StreamingDataset(IterableDataset):
 
     def __iter__(self) -> "StreamingDataset":
         # When the StreamingDataset is used within map or optimize, let's refetch the distributed env.
+        logger.debug(_get_log_msg({"name": "iterating_dataset", "ph": "B"}))
         if os.getenv("DATA_OPTIMIZER_GLOBAL_RANK"):
             self.distributed_env = _DistributedEnv.detect()
 
         self.worker_env = _WorkerEnv.detect()
         self.cache = self._create_cache(worker_env=self.worker_env)
         self.shuffler = self._create_shuffler(self.cache)
+        self.on_demand_bytes = False  # reset on_demand_bytes to False, and store chunks in the cache
 
         # Handle restart
         if self._state_dict:
             self._validate_state_dict()
-            state: Dict[str, Any] = self._state_dict
+            state: dict[str, Any] = self._state_dict
             self.current_epoch = state["current_epoch"]
 
         workers_chunks, workers_intervals = self.shuffler.get_chunks_and_intervals_per_workers(
@@ -309,22 +359,22 @@ class StreamingDataset(IterableDataset):
                 ]
 
             self.num_chunks = len(self.worker_chunks)
-            self.current_indexes = []
-            self.chunk_index = 0
+            self.upcoming_indexes = []
+            self.worker_next_chunk_index = 0
             self.global_index = 0
-            self.index = 0
+            self.consumed_sample_count_in_curr_chunk = 0
 
         self.has_triggered_download = False
         self.last_time = time()
 
         return self
 
-    def _resume(self, workers_chunks: List[List[int]], workers_intervals: List[Any]) -> None:
+    def _resume(self, workers_chunks: list[list[int]], workers_intervals: list[Any]) -> None:
         assert self._state_dict
         assert self.worker_env
         assert self.shuffler
 
-        state: Dict[str, Any] = self._state_dict
+        state: dict[str, Any] = self._state_dict
 
         num_workers = state["num_workers"]
         batch_size = state["batch_size"]
@@ -347,27 +397,35 @@ class StreamingDataset(IterableDataset):
         worker_local_rank = self.worker_env.rank
 
         self.num_chunks = len(workers_intervals[worker_rank])
-        self.chunk_index = chunks_index[worker_local_rank]
+        self.worker_next_chunk_index = chunks_index[worker_local_rank]
         self.worker_chunks = workers_chunks[worker_rank]
         self.worker_intervals = workers_intervals[worker_rank]
 
+        if self.worker_next_chunk_index >= self.num_chunks:
+            # This can happen when interrupting and resuming after some but not all workers are done.
+            # Proceeding would result in an indexing error when attempting to access the next chunk.
+            # To prevent this we exit early and let the worker raise a StopIteration in __next__.
+            return
+
         # replay the indexes for the current chunks
-        interval = self.worker_intervals[self.chunk_index]
+        interval = self.worker_intervals[self.worker_next_chunk_index]
         current_indexes = np.arange(interval[1], interval[2])
 
         # re-shuffle the indexes
-        current_indexes = self.shuffler(current_indexes, self.num_chunks, self.current_epoch, self.chunk_index)
+        current_indexes = self.shuffler(
+            current_indexes, self.num_chunks, self.current_epoch, self.worker_next_chunk_index
+        )
 
         # skip any indexes already consumed
         current_indexes = current_indexes[indexes[worker_local_rank] :]
-        self.current_indexes = current_indexes
+        self.upcoming_indexes = current_indexes
 
         self.global_index = indexes[worker_local_rank]
 
         # bump the chunk_index
-        self.chunk_index += 1
+        self.worker_next_chunk_index += 1
 
-    def __getitem__(self, index: Union[ChunkedIndex, int]) -> Any:
+    def __getitem__(self, index: Union[ChunkedIndex, int, slice]) -> Any:
         if self.cache is None:
             self.worker_env = _WorkerEnv.detect()
             self.cache = self._create_cache(worker_env=self.worker_env)
@@ -375,64 +433,101 @@ class StreamingDataset(IterableDataset):
         if isinstance(index, int):
             index = ChunkedIndex(*self.cache._get_chunk_index_from_index(index))
         elif isinstance(index, slice):
+            self.on_demand_bytes = False  # for slices, we always want to store the chunks
             start, stop, step = index.indices(len(self))
             _my_indices = list(range(start, stop, step))
             _my_cache_indices = [ChunkedIndex(*self.cache._get_chunk_index_from_index(idx)) for idx in _my_indices]
             return [self.cache[chnk_idx] for chnk_idx in _my_cache_indices]
-        return self.cache[index]
+        logger.debug(
+            _get_log_msg(
+                {"name": f"getitem_dataset_for_chunk_index_{index.chunk_index}_and_index_{index.index}", "ph": "B"}
+            )
+        )
+        item = self.cache[index]
+        logger.debug(
+            _get_log_msg(
+                {"name": f"getitem_dataset_for_chunk_index_{index.chunk_index}_and_index_{index.index}", "ph": "E"}
+            )
+        )
+        if hasattr(self, "transform"):
+            if isinstance(self.transform, list):
+                for transform_fn in self.transform:
+                    item = transform_fn(item)
+            else:
+                item = self.transform(item)
+
+        return item
 
     def __next__(self) -> Any:
-        # Prevent to create more batch on a given process
+        # check if we have reached the end of the dataset (i.e., all the chunks have been processed)
         if self.global_index >= self.stop_length:
+            # global_index: total number of samples processed by the current worker across all chunks
+            # stop_length: max number of samples that the current worker will process
+            # if they are equal, means, worker has processed all the chunks
             self.current_epoch += 1
             self.reset_state_dict()
+            logger.debug(_get_log_msg({"name": "iterating_dataset", "ph": "E"}))
+            self.on_demand_bytes = True  # reset on_demand_bytes to True
             raise StopIteration
 
         # Lazily re-populate the interval to reduce memory usage.
-        if len(self.current_indexes) == 0:
-            if self.chunk_index == self.num_chunks:
+        if len(self.upcoming_indexes) == 0:
+            # check if it's not the end of the epoch
+            if self.num_chunks is not None and self.worker_next_chunk_index >= self.num_chunks:
                 self.current_epoch += 1
                 self.reset_state_dict()
+                self.on_demand_bytes = True  # reset on_demand_bytes to True
                 raise StopIteration
 
-            # reset index
-            self.index = 0
+            # if upcoming_indexes is empty, means either:
+            #   - it's the start of a new epoch, or,
+            #   - we have processed all the indexes in the current chunk
+            #
+            # we need to move to the next chunk (or first chunk if it's the start)
+            # reset consumed_sample_count_in_curr_chunk as we are switching to a new chunk
+            self.consumed_sample_count_in_curr_chunk = 0
 
-            interval = self.worker_intervals[self.chunk_index]
+            # `next_worker_chunks_index` is the index of the chunk that we will be working on now
+            interval = self.worker_intervals[self.worker_next_chunk_index]
             current_indexes = np.arange(interval[1], interval[2])
 
             assert self.shuffler is not None
             assert self.num_chunks is not None
-            self.current_indexes = self.shuffler(current_indexes, self.num_chunks, self.current_epoch, self.chunk_index)
+            self.upcoming_indexes = self.shuffler(
+                current_indexes, self.num_chunks, self.current_epoch, self.worker_next_chunk_index
+            )
 
-            self.chunk_index += 1
+            self.worker_next_chunk_index += 1  # bump the chunk_index
 
         # Get the first index
-        index = self.current_indexes.pop(0)
+        index = self.upcoming_indexes.pop(0)
 
         # Call the `__getitem__` method.
         data = self.__getitem__(
             ChunkedIndex(
                 index=index,
-                chunk_index=self.worker_chunks[self.chunk_index - 1],
+                chunk_index=self.worker_chunks[self.worker_next_chunk_index - 1],
                 # We provide the chunks indexes only one the first
-                chunk_indexes=None if self.has_triggered_download else self.worker_chunks[self.chunk_index - 1 :],
-                is_last_index=(self.chunk_index - 1) == len(self.worker_intervals) and len(self.current_indexes) == 1,
+                chunk_indexes=None
+                if self.has_triggered_download
+                else self.worker_chunks[self.worker_next_chunk_index - 1 :],
+                is_last_index=(self.worker_next_chunk_index) == self.num_chunks and len(self.upcoming_indexes) == 0,
             )
         )
 
         self.has_triggered_download = True
-        self.global_index += 1
-        self.index += 1
+        self.global_index += 1  # total number of samples processed by the current worker
+        self.consumed_sample_count_in_curr_chunk += 1  # number of samples processed in the current chunk
 
         return data
 
-    def state_dict(self, num_samples_yielded: int, num_workers: int, batch_size: int) -> Dict[str, Any]:
+    def state_dict(self, num_samples_yielded: int, num_workers: int, batch_size: int) -> dict[str, Any]:
         if _is_in_dataloader_worker():
             raise RuntimeError("The method `state_dict` should only be called in the main process.")
 
         if self._state_dict is not None:
             self._state_dict["num_samples_yielded"] = num_samples_yielded
+            self._state_dict["current_epoch"] = self.current_epoch
             return self._state_dict
 
         return {
@@ -452,7 +547,7 @@ class StreamingDataset(IterableDataset):
             "region_of_interest": self.region_of_interest,
         }
 
-    def load_state_dict(self, state_dict: Dict[str, Any]) -> None:
+    def load_state_dict(self, state_dict: dict[str, Any]) -> None:
         if state_dict:
             # the state is restored within the workers
             self._state_dict = state_dict
@@ -470,7 +565,7 @@ class StreamingDataset(IterableDataset):
         assert self.worker_env
         assert self.cache
 
-        state: Dict[str, Any] = self._state_dict
+        state: dict[str, Any] = self._state_dict
         if state["shuffle"] != self.shuffle:
             if not self._force_override_state_dict:
                 raise ValueError(
@@ -576,7 +671,7 @@ class StreamingDataset(IterableDataset):
 
     def reset(self) -> None:
         # undo all the properties associated with original dataset
-        default_properties: Dict[str, Any] = {
+        default_properties: dict[str, Any] = {
             "cache": None,
             "worker_env": None,
             "worker_chunks": [],
@@ -608,7 +703,7 @@ def is_integer(value: str) -> bool:
         return False
 
 
-def _replay_sampling(num_samples_yielded: int, batch_size: int, num_workers: int) -> Dict[int, int]:
+def _replay_sampling(num_samples_yielded: int, batch_size: int, num_workers: int) -> dict[int, int]:
     """This function replays the sampling from the dataloader."""
     divisible_num_batches_yielded = num_samples_yielded // (num_workers * batch_size)
 
@@ -632,8 +727,8 @@ def _replay_sampling(num_samples_yielded: int, batch_size: int, num_workers: int
 
 
 def _replay_chunks_sampling(
-    workers_intervals: Dict[int, List[Any]], indexes: Dict[int, int]
-) -> Tuple[Dict[int, int], Dict[int, int]]:
+    workers_intervals: dict[int, list[Any]], indexes: dict[int, int]
+) -> tuple[dict[int, int], dict[int, int]]:
     chunks_index = {}
 
     for worker_idx in range(len(workers_intervals)):
