@@ -19,7 +19,7 @@ from copy import deepcopy
 from io import BytesIO, FileIO
 from multiprocessing import Queue
 from time import sleep, time
-from typing import Any, Dict, List, Optional, Tuple, Union
+from typing import Any, Optional, Union
 
 import numpy as np
 import torch
@@ -32,13 +32,14 @@ from litdata.constants import (
     _PYARROW_AVAILABLE,
     _TORCH_DTYPES_MAPPING,
 )
+from litdata.debugger import ChromeTraceColors, _get_log_msg
 from litdata.streaming.serializers import Serializer
 from litdata.utilities._pytree import PyTree, tree_unflatten
 from litdata.utilities.encryption import Encryption, EncryptionLevel
 
 Interval = namedtuple("Interval", ["chunk_start", "roi_start_idx", "roi_end_idx", "chunk_end"])
 
-logger = logging.getLogger(__name__)
+logger = logging.getLogger("litdata.streaming.item_loader")
 
 
 class BaseItemLoader(ABC):
@@ -46,10 +47,10 @@ class BaseItemLoader(ABC):
 
     def setup(
         self,
-        config: Dict,
-        chunks: List,
-        serializers: Dict[str, Serializer],
-        region_of_interest: Optional[List[Tuple[int, int]]] = None,
+        config: dict,
+        chunks: list,
+        serializers: dict[str, Serializer],
+        region_of_interest: Optional[list[tuple[int, int]]] = None,
         force_download_queue: Optional[Queue] = None,
     ) -> None:
         self._config = config
@@ -79,11 +80,11 @@ class BaseItemLoader(ABC):
             return serialier
         return data_format
 
-    def state_dict(self) -> Dict:
+    def state_dict(self) -> dict:
         return {}
 
     @abstractmethod
-    def generate_intervals(self) -> List[Interval]:
+    def generate_intervals(self) -> list[Interval]:
         """Returns a list of intervals.
 
         The structure is: [chunk_start, region_of_interest_start, region_of_interest_end, chunk_end]
@@ -106,12 +107,20 @@ class BaseItemLoader(ABC):
     ) -> Any:
         """Returns an item loaded from a chunk."""
 
+    def load_item_from_bytes(
+        self,
+        raw_bytes: bytes,
+        chunk_index: int,
+    ) -> Any:
+        """Returns an item loaded from bytes."""
+        raise NotImplementedError("The `load_item_from_bytes` method is not implemented for this item loader.")
+
     @abstractmethod
     def delete(self, chunk_index: int, chunk_filepath: str) -> None:
         """Delete a chunk from the local filesystem."""
 
     @abstractmethod
-    def encode_data(self, data: List[bytes], sizes: List[int], flattened: List[Any]) -> Any:
+    def encode_data(self, data: list[bytes], sizes: list[int], flattened: list[Any]) -> Any:
         pass
 
 
@@ -121,10 +130,10 @@ class PyTreeLoader(BaseItemLoader):
     def __init__(self) -> None:
         super().__init__()
         self._chunk_filepath: str | None = None
-        self._decrypted_chunks: Dict[int, bytes] = {}
+        self._decrypted_chunks: dict[int, bytes] = {}
         self._open_handle: FileIO | None = None
 
-    def generate_intervals(self) -> List[Interval]:
+    def generate_intervals(self) -> list[Interval]:
         intervals = []
         begin = 0
         end = 0
@@ -142,6 +151,22 @@ class PyTreeLoader(BaseItemLoader):
     def pre_load_chunk(self, chunk_index: int, chunk_filepath: str) -> None:
         pass
 
+    def load_item_from_bytes(
+        self,
+        raw_bytes: bytes,
+        chunk_index: int,
+    ) -> bytes:
+        if self._config.get("encryption"):
+            raise ValueError("The `load_item_from_bytes` method does not support encrypted data loading currently.")
+
+        # check for mosaic mds format
+        if "format" in self._config and self._config["format"] == "mds":
+            item_data = self.mds_deserialize(raw_bytes, chunk_index)
+        else:
+            item_data = self.deserialize(raw_bytes)
+
+        return item_data
+
     def load_item_from_chunk(
         self,
         index: int,
@@ -172,6 +197,9 @@ class PyTreeLoader(BaseItemLoader):
         #       => 3 * 4 = 12          # each takes 4 bytes
         #       => offset = 12
         #
+        logger.debug(
+            _get_log_msg({"name": f"load_item_from_chunk_for_chunk_index_{chunk_index}_and_index_{index}", "ph": "B"})
+        )
         offset = (1 + (index - begin) if index >= begin else index + 1) * 4
 
         if chunk_filepath != self._chunk_filepath:
@@ -206,9 +234,15 @@ class PyTreeLoader(BaseItemLoader):
 
         # check for mosaic mds format
         if "format" in self._config and self._config["format"] == "mds":
-            return self.mds_deserialize(data, chunk_index)
+            item_data = self.mds_deserialize(data, chunk_index)
+        else:
+            item_data = self.deserialize(data)
 
-        return self.deserialize(data)
+        logger.debug(
+            _get_log_msg({"name": f"load_item_from_chunk_for_chunk_index_{chunk_index}_and_index_{index}", "ph": "E"})
+        )
+
+        return item_data
 
     def _load_encrypted_data(
         self, chunk_filepath: str, chunk_index: int, offset: int, encryption: Optional[Encryption]
@@ -293,8 +327,26 @@ class PyTreeLoader(BaseItemLoader):
             self._open_handle = None
 
     def delete(self, chunk_index: int, chunk_filepath: str) -> None:
+        logger.debug(
+            _get_log_msg(
+                {
+                    "name": f"delete_chunk_for_chunk_index_{chunk_index}",
+                    "ph": "B",
+                    "cname": ChromeTraceColors.BRIGHT_RED,
+                }
+            )
+        )
         if os.path.exists(chunk_filepath):
             os.remove(chunk_filepath)
+        logger.debug(
+            _get_log_msg(
+                {
+                    "name": f"delete_chunk_for_chunk_index_{chunk_index}",
+                    "ph": "E",
+                    "cname": ChromeTraceColors.BRIGHT_RED,
+                }
+            )
+        )
 
     def _validate_encryption(self, encryption: Optional[Encryption]) -> None:
         """Validate the encryption object."""
@@ -306,7 +358,7 @@ class PyTreeLoader(BaseItemLoader):
             raise ValueError("Encryption level mismatch.")
 
     @classmethod
-    def encode_data(cls, data: List[bytes], sizes: List[int], flattened: List[Any]) -> Tuple[bytes, Optional[int]]:
+    def encode_data(cls, data: list[bytes], sizes: list[int], flattened: list[Any]) -> tuple[bytes, Optional[int]]:
         """Encodes multiple serialized objects into a single binary format with size metadata.
 
         This method combines multiple serialized objects into a single byte array, prefixed with their sizes.
@@ -352,14 +404,14 @@ class TokensLoader(BaseItemLoader):
         """
         super().__init__()
         self._block_size = block_size
-        self._mmaps: Dict[int, np.memmap] = {}
-        self._buffers: Dict[int, bytes] = {}
+        self._mmaps: dict[int, np.memmap] = {}
+        self._buffers: dict[int, bytes] = {}
         # keeps track of number of readers for each chunk (can be more than 1 if multiple workers are reading)
         self._counter = defaultdict(int)
         self._dtype: Optional[torch.dtype] = None
-        self._chunk_filepaths: Dict[str, bool] = {}
+        self._chunk_filepaths: dict[str, bool] = {}
 
-    def state_dict(self) -> Dict:
+    def state_dict(self) -> dict:
         assert self._block_size
         return {
             "block_size": self._block_size,
@@ -367,10 +419,10 @@ class TokensLoader(BaseItemLoader):
 
     def setup(
         self,
-        config: Dict,
-        chunks: List,
-        serializers: Dict[str, Serializer],
-        region_of_interest: Optional[List[Tuple[int, int]]] = None,
+        config: dict,
+        chunks: list,
+        serializers: dict[str, Serializer],
+        region_of_interest: Optional[list[tuple[int, int]]] = None,
     ) -> None:
         super().setup(config, chunks, serializers, region_of_interest)
 
@@ -387,7 +439,7 @@ class TokensLoader(BaseItemLoader):
         if all(chunk["dim"] is None for chunk in self._chunks):
             raise ValueError("The provided chunks isn't properly setup.")
 
-    def generate_intervals(self) -> List[Interval]:
+    def generate_intervals(self) -> list[Interval]:
         assert self._block_size
         intervals = []
         begin = 0
@@ -440,6 +492,10 @@ class TokensLoader(BaseItemLoader):
         if chunk_filepath in self._chunk_filepaths and not os.path.isfile(chunk_filepath):
             del self._chunk_filepaths[chunk_filepath]
 
+        logger.debug(
+            _get_log_msg({"name": f"load_item_from_chunk_for_chunk_index_{chunk_index}_and_index_{index}", "ph": "B"})
+        )
+
         if chunk_filepath not in self._chunk_filepaths:
             exists = os.path.exists(chunk_filepath) and os.stat(chunk_filepath).st_size > filesize_bytes
 
@@ -479,9 +535,22 @@ class TokensLoader(BaseItemLoader):
         else:
             # count: number of tokens to read from buffer => `self._block_size`
             data = np.frombuffer(buffer, dtype=self._dtype, count=self._block_size, offset=offset)  # type: ignore
+
+        logger.debug(
+            _get_log_msg({"name": f"load_item_from_chunk_for_chunk_index_{chunk_index}_and_index_{index}", "ph": "E"})
+        )
         return data
 
     def delete(self, chunk_index: int, chunk_filepath: str) -> None:
+        logger.debug(
+            _get_log_msg(
+                {
+                    "name": f"delete_chunk_for_chunk_index_{chunk_index}",
+                    "ph": "B",
+                    "cname": ChromeTraceColors.BRIGHT_RED,
+                }
+            )
+        )
         if os.path.exists(chunk_filepath):
             if chunk_index in self._buffers:
                 del self._buffers[chunk_index]
@@ -491,6 +560,15 @@ class TokensLoader(BaseItemLoader):
                 del self._mmaps[chunk_index]
                 del self._counter[chunk_index]
             os.remove(chunk_filepath)
+        logger.debug(
+            _get_log_msg(
+                {
+                    "name": f"delete_chunk_for_chunk_index_{chunk_index}",
+                    "ph": "E",
+                    "cname": ChromeTraceColors.BRIGHT_RED,
+                }
+            )
+        )
 
     def close(self, chunk_index: int) -> None:
         """Release the memory-mapped file for a specific chunk index."""
@@ -504,7 +582,7 @@ class TokensLoader(BaseItemLoader):
                 del self._mmaps[chunk_index]
 
     @classmethod
-    def encode_data(cls, data: List[bytes], _: List[int], flattened: List[Any]) -> Tuple[bytes, Optional[int]]:
+    def encode_data(cls, data: list[bytes], _: list[int], flattened: list[Any]) -> tuple[bytes, Optional[int]]:
         r"""Encodes tokenized data into a raw byte format while preserving dimensional information.
 
         Parameters:
@@ -539,7 +617,7 @@ class ParquetLoader(BaseItemLoader):
         if not _PYARROW_AVAILABLE:
             raise ModuleNotFoundError("Please, run: `pip install pyarrow`")
 
-        self._chunk_filepaths: Dict[str, bool] = {}
+        self._chunk_filepaths: dict[str, bool] = {}
         self._pre_load_chunk = pre_load_chunk
         self._low_memory = low_memory
 
@@ -552,10 +630,10 @@ class ParquetLoader(BaseItemLoader):
 
     def setup(
         self,
-        config: Dict,
-        chunks: List,
-        serializers: Dict[str, Serializer],
-        region_of_interest: Optional[List[Tuple[int, int]]] = None,
+        config: dict,
+        chunks: list,
+        serializers: dict[str, Serializer],
+        region_of_interest: Optional[list[tuple[int, int]]] = None,
     ) -> None:
         self._config = config
         self._chunks = chunks
@@ -563,11 +641,11 @@ class ParquetLoader(BaseItemLoader):
         self._data_format = self._config["data_format"]
         self._shift_idx = len(self._data_format) * 4
         self.region_of_interest = region_of_interest
-        self._df: Dict[int, Any] = {}
-        self._chunk_row_groups: Dict[int, Any] = {}
-        self._chunk_row_group_item_read_count: Dict[int, Any] = {}
+        self._df: dict[int, Any] = {}
+        self._chunk_row_groups: dict[int, Any] = {}
+        self._chunk_row_group_item_read_count: dict[int, Any] = {}
 
-    def generate_intervals(self) -> List[Interval]:
+    def generate_intervals(self) -> list[Interval]:
         intervals = []
         begin = 0
         end = 0
@@ -604,6 +682,9 @@ class ParquetLoader(BaseItemLoader):
         if chunk_filepath in self._chunk_filepaths and not os.path.isfile(chunk_filepath):
             del self._chunk_filepaths[chunk_filepath]
 
+        logger.debug(
+            _get_log_msg({"name": f"load_item_from_chunk_for_chunk_index_{chunk_index}_and_index_{index}", "ph": "B"})
+        )
         if chunk_filepath not in self._chunk_filepaths:
             exists = os.path.exists(chunk_filepath) and os.stat(chunk_filepath).st_size >= filesize_bytes
 
@@ -616,9 +697,14 @@ class ParquetLoader(BaseItemLoader):
         # relative index of the desired row within the chunk.
         relative_index = index - begin
         if self._low_memory:
-            return self._get_item_with_low_memory(chunk_index, chunk_filepath, relative_index)
+            item_data = self._get_item_with_low_memory(chunk_index, chunk_filepath, relative_index)
+        else:
+            item_data = self._get_item(chunk_index, chunk_filepath, relative_index)
 
-        return self._get_item(chunk_index, chunk_filepath, relative_index)
+        logger.debug(
+            _get_log_msg({"name": f"load_item_from_chunk_for_chunk_index_{chunk_index}_and_index_{index}", "ph": "E"})
+        )
+        return item_data
 
     def _get_item_with_low_memory(self, chunk_index: int, chunk_filepath: str, row_index: int) -> Any:
         """Retrieve a dataframe row from a parquet chunk in low memory mode.
@@ -703,6 +789,15 @@ class ParquetLoader(BaseItemLoader):
 
     def delete(self, chunk_index: int, chunk_filepath: str) -> None:
         """Delete a chunk from the local filesystem."""
+        logger.debug(
+            _get_log_msg(
+                {
+                    "name": f"delete_chunk_for_chunk_index_{chunk_index}",
+                    "ph": "B",
+                    "cname": ChromeTraceColors.BRIGHT_RED,
+                }
+            )
+        )
         if chunk_index in self._df:
             del self._df[chunk_index]
         if chunk_index in self._chunk_row_groups:
@@ -712,6 +807,15 @@ class ParquetLoader(BaseItemLoader):
             del self._chunk_row_group_item_read_count[chunk_index]
         if os.path.exists(chunk_filepath):
             os.remove(chunk_filepath)
+        logger.debug(
+            _get_log_msg(
+                {
+                    "name": f"delete_chunk_for_chunk_index_{chunk_index}",
+                    "ph": "E",
+                    "cname": ChromeTraceColors.BRIGHT_RED,
+                }
+            )
+        )
 
     def close(self, chunk_index: int) -> None:
         """Release the memory-mapped file for a specific chunk index."""
@@ -724,5 +828,5 @@ class ParquetLoader(BaseItemLoader):
         if chunk_index in self._chunk_row_group_item_read_count:
             del self._chunk_row_group_item_read_count[chunk_index]
 
-    def encode_data(self, data: List[bytes], sizes: List[int], flattened: List[Any]) -> Any:
+    def encode_data(self, data: list[bytes], sizes: list[int], flattened: list[Any]) -> Any:
         pass
