@@ -19,20 +19,19 @@ from abc import ABC, abstractmethod
 from collections import OrderedDict
 from contextlib import suppress
 from copy import deepcopy
-from typing import TYPE_CHECKING, Any, Dict, Optional, Tuple, Union
+from itertools import chain
+from typing import Any, Optional
 
 import numpy as np
 import tifffile
 import torch
-from lightning_utilities.core.imports import RequirementCache
 
-from litdata.constants import _NUMPY_DTYPES_MAPPING, _TORCH_DTYPES_MAPPING
-
-if TYPE_CHECKING:
-    from PIL.JpegImagePlugin import JpegImageFile
-_PIL_AVAILABLE = RequirementCache("PIL")
-_TORCH_VISION_AVAILABLE = RequirementCache("torchvision")
-_AV_AVAILABLE = RequirementCache("av")
+from litdata.constants import (
+    _AV_AVAILABLE,
+    _NUMPY_DTYPES_MAPPING,
+    _PIL_AVAILABLE,
+    _TORCH_DTYPES_MAPPING,
+)
 
 
 class Serializer(ABC):
@@ -43,7 +42,7 @@ class Serializer(ABC):
     """
 
     @abstractmethod
-    def serialize(self, data: Any) -> Tuple[bytes, Optional[str]]:
+    def serialize(self, data: Any) -> tuple[bytes, Optional[str]]:
         pass
 
     @abstractmethod
@@ -61,7 +60,7 @@ class Serializer(ABC):
 class PILSerializer(Serializer):
     """The PILSerializer serialize and deserialize PIL Image to and from bytes."""
 
-    def serialize(self, item: Any) -> Tuple[bytes, Optional[str]]:
+    def serialize(self, item: Any) -> tuple[bytes, Optional[str]]:
         mode = item.mode.encode("utf-8")
         width, height = item.size
         raw = item.tobytes()
@@ -95,7 +94,7 @@ class PILSerializer(Serializer):
 class JPEGSerializer(Serializer):
     """The JPEGSerializer serialize and deserialize JPEG image to and from bytes."""
 
-    def serialize(self, item: Any) -> Tuple[bytes, Optional[str]]:
+    def serialize(self, item: Any) -> tuple[bytes, Optional[str]]:
         if not _PIL_AVAILABLE:
             raise ModuleNotFoundError("PIL is required. Run `pip install pillow`")
 
@@ -129,20 +128,17 @@ class JPEGSerializer(Serializer):
 
         raise TypeError(f"The provided item should be of type `JpegImageFile`. Found {item}.")
 
-    def deserialize(self, data: bytes) -> Union["JpegImageFile", torch.Tensor]:
-        if _TORCH_VISION_AVAILABLE:
-            from torchvision.io import decode_jpeg
-            from torchvision.transforms.functional import pil_to_tensor
+    def deserialize(self, data: bytes) -> torch.Tensor:
+        from torchvision.io import decode_image, decode_jpeg
 
-            array = torch.frombuffer(data, dtype=torch.uint8)
-            # Note: Some datasets like Imagenet contains some PNG images with JPEG extension, so we fallback to PIL
-            with suppress(RuntimeError):
-                return decode_jpeg(array)
+        array = torch.frombuffer(data, dtype=torch.uint8)
+        # Try decoding as JPEG. Some datasets (e.g., ImageNet) may have PNG images with a JPEG extension,
+        # which will cause decode_jpeg to fail. In that case, fall back to a generic image decoder.
+        with suppress(RuntimeError):
+            return decode_jpeg(array)
 
-        img = PILSerializer.deserialize(data)
-        if _TORCH_VISION_AVAILABLE:
-            img = pil_to_tensor(img)
-        return img
+        # Fallback: decode as a generic image (handles PNG, etc.)
+        return decode_image(array)
 
     def can_serialize(self, item: Any) -> bool:
         if not _PIL_AVAILABLE:
@@ -153,10 +149,79 @@ class JPEGSerializer(Serializer):
         return isinstance(item, JpegImageFile)
 
 
+class JPEGArraySerializer(Serializer):
+    """The JPEGArraySerializer serializes and deserializes lists of JPEG images to and from bytes."""
+
+    def serialize(self, item: Any) -> tuple[bytes, Optional[str]]:
+        # Store number of images as first 4 bytes
+        n_images_bytes = np.uint32(len(item)).tobytes()
+
+        # create a instance of JPEGSerializer
+        if not hasattr(self, "_jpeg_serializer"):
+            self._jpeg_serializer = JPEGSerializer()
+
+        # convert each image to bytes and store in a list
+        image_bytes = []
+        for image in item:
+            image_bytes.append(self._jpeg_serializer.serialize(image)[0])
+
+        # Store all image sizes as uint32 array and convert to bytes
+        image_sizes_bytes = np.array([len(elem) for elem in image_bytes], dtype=np.uint32).tobytes()
+
+        # Concatenate all data: n_images + sizes + image bytes
+        return b"".join(chain([n_images_bytes, image_sizes_bytes], image_bytes)), None
+
+    def deserialize(self, data: bytes) -> list[torch.Tensor]:
+        if len(data) < 4:
+            raise ValueError("Input data is too short to contain valid list of images")
+
+        # Extract number of images from the first 4 bytes
+        n_images = np.frombuffer(data[:4], dtype=np.uint32)[0]
+
+        # Ensure the number of images is positive
+        if n_images <= 0:
+            raise ValueError("Number of images must be positive")
+
+        # Calculate the offset where image bytes start
+        image_bytes_offset = 4 + 4 * n_images
+
+        if len(data) < image_bytes_offset:
+            raise ValueError("Data is too short for the number of images specified")
+
+        # Extract the sizes of each image
+        image_sizes = np.frombuffer(data[4:image_bytes_offset], dtype=np.uint32)
+
+        # Calculate offsets for each image's data
+        offsets = np.cumsum(np.concatenate(([image_bytes_offset], image_sizes)))
+
+        if len(offsets) != n_images + 1:
+            raise ValueError("Mismatch between number of images and offsets")
+
+        if not hasattr(self, "_jpeg_serializer"):
+            self._jpeg_serializer = JPEGSerializer()
+
+        # Extract and decode each image data
+        images = []
+        for i in range(n_images):
+            # Extract the image data using the offsets
+            image_data = data[offsets[i] : offsets[i + 1]]
+            # Convert the image data to a tensor
+            images.append(self._jpeg_serializer.deserialize(image_data))
+        return images
+
+    def can_serialize(self, item: Any) -> bool:
+        """Check if the item is a list of JPEG images."""
+        if not _PIL_AVAILABLE:
+            return False
+        from PIL.JpegImagePlugin import JpegImageFile
+
+        return isinstance(item, (list, tuple)) and all(isinstance(elem, JpegImageFile) for elem in item)
+
+
 class BytesSerializer(Serializer):
     """The BytesSerializer serialize and deserialize integer to and from bytes."""
 
-    def serialize(self, item: bytes) -> Tuple[bytes, Optional[str]]:
+    def serialize(self, item: bytes) -> tuple[bytes, Optional[str]]:
         return item, None
 
     def deserialize(self, item: bytes) -> bytes:
@@ -173,7 +238,7 @@ class TensorSerializer(Serializer):
         super().__init__()
         self._dtype_to_indices = {v: k for k, v in _TORCH_DTYPES_MAPPING.items()}
 
-    def serialize(self, item: torch.Tensor) -> Tuple[bytes, Optional[str]]:
+    def serialize(self, item: torch.Tensor) -> tuple[bytes, Optional[str]]:
         dtype_indice = self._dtype_to_indices[item.dtype]
         data = [np.uint32(dtype_indice).tobytes()]
         data.append(np.uint32(len(item.shape)).tobytes())
@@ -216,7 +281,7 @@ class NoHeaderTensorSerializer(Serializer):
     def setup(self, data_format: str) -> None:
         self._dtype = _TORCH_DTYPES_MAPPING[int(data_format.split(":")[1])]
 
-    def serialize(self, item: torch.Tensor) -> Tuple[bytes, Optional[str]]:
+    def serialize(self, item: torch.Tensor) -> tuple[bytes, Optional[str]]:
         dtype_indice = self._dtype_to_indices[item.dtype]
         return item.numpy().tobytes(order="C"), f"no_header_tensor:{dtype_indice}"
 
@@ -235,7 +300,7 @@ class NumpySerializer(Serializer):
         super().__init__()
         self._dtype_to_indices = {v: k for k, v in _NUMPY_DTYPES_MAPPING.items()}
 
-    def serialize(self, item: np.ndarray) -> Tuple[bytes, Optional[str]]:
+    def serialize(self, item: np.ndarray) -> tuple[bytes, Optional[str]]:
         dtype_indice = self._dtype_to_indices[item.dtype]
         data = [np.uint32(dtype_indice).tobytes()]
         data.append(np.uint32(len(item.shape)).tobytes())
@@ -275,7 +340,7 @@ class NoHeaderNumpySerializer(Serializer):
     def setup(self, data_format: str) -> None:
         self._dtype = _NUMPY_DTYPES_MAPPING[int(data_format.split(":")[1])]
 
-    def serialize(self, item: np.ndarray) -> Tuple[bytes, Optional[str]]:
+    def serialize(self, item: np.ndarray) -> tuple[bytes, Optional[str]]:
         dtype_indice: int = self._dtype_to_indices[item.dtype]
         return item.tobytes(order="C"), f"no_header_numpy:{dtype_indice}"
 
@@ -290,7 +355,7 @@ class NoHeaderNumpySerializer(Serializer):
 class PickleSerializer(Serializer):
     """The PickleSerializer serialize and deserialize python objects to and from bytes."""
 
-    def serialize(self, item: Any) -> Tuple[bytes, Optional[str]]:
+    def serialize(self, item: Any) -> tuple[bytes, Optional[str]]:
         return pickle.dumps(item), None
 
     def deserialize(self, data: bytes) -> Any:
@@ -301,7 +366,7 @@ class PickleSerializer(Serializer):
 
 
 class FileSerializer(Serializer):
-    def serialize(self, filepath: str) -> Tuple[bytes, Optional[str]]:
+    def serialize(self, filepath: str) -> tuple[bytes, Optional[str]]:
         print("FileSerializer will be removed in the future.")
         _, file_extension = os.path.splitext(filepath)
         with open(filepath, "rb") as f:
@@ -320,16 +385,13 @@ class FileSerializer(Serializer):
 class VideoSerializer(Serializer):
     _EXTENSIONS = ("mp4", "ogv", "mjpeg", "avi", "mov", "h264", "mpg", "webm", "wmv")
 
-    def serialize(self, filepath: str) -> Tuple[bytes, Optional[str]]:
+    def serialize(self, filepath: str) -> tuple[bytes, Optional[str]]:
         _, file_extension = os.path.splitext(filepath)
         with open(filepath, "rb") as f:
             file_extension = file_extension.replace(".", "").lower()
             return f.read(), f"video:{file_extension}"
 
     def deserialize(self, data: bytes) -> Any:
-        if not _TORCH_VISION_AVAILABLE:
-            raise ModuleNotFoundError("torchvision is required. Run `pip install torchvision`")
-
         if not _AV_AVAILABLE:
             raise ModuleNotFoundError("av is required. Run `pip install av`")
 
@@ -348,7 +410,7 @@ class VideoSerializer(Serializer):
 
 
 class StringSerializer(Serializer):
-    def serialize(self, obj: str) -> Tuple[bytes, Optional[str]]:
+    def serialize(self, obj: str) -> tuple[bytes, Optional[str]]:
         return obj.encode("utf-8"), None
 
     def deserialize(self, data: bytes) -> str:
@@ -365,7 +427,7 @@ class NumericSerializer:
         self.dtype = dtype
         self.size = self.dtype().nbytes
 
-    def serialize(self, obj: Any) -> Tuple[bytes, Optional[str]]:
+    def serialize(self, obj: Any) -> tuple[bytes, Optional[str]]:
         return self.dtype(obj).tobytes(), None
 
     def deserialize(self, data: bytes) -> Any:
@@ -391,7 +453,7 @@ class FloatSerializer(NumericSerializer, Serializer):
 class BooleanSerializer(Serializer):
     """The BooleanSerializer serializes and deserializes boolean values to and from bytes."""
 
-    def serialize(self, item: bool) -> Tuple[bytes, Optional[str]]:
+    def serialize(self, item: bool) -> tuple[bytes, Optional[str]]:
         """Serialize a boolean value to bytes.
 
         Args:
@@ -428,7 +490,7 @@ class BooleanSerializer(Serializer):
 class TIFFSerializer(Serializer):
     """Serializer for TIFF files using tifffile."""
 
-    def serialize(self, item: Any) -> Tuple[bytes, Optional[str]]:
+    def serialize(self, item: Any) -> tuple[bytes, Optional[str]]:
         if not isinstance(item, str) or not os.path.isfile(item):
             raise ValueError(f"The item to serialize must be a valid file path. Received: {item}")
 
@@ -456,6 +518,7 @@ _SERIALIZERS = OrderedDict(
         "file": FileSerializer(),
         "pil": PILSerializer(),
         "jpeg": JPEGSerializer(),
+        "jpeg_array": JPEGArraySerializer(),
         "bytes": BytesSerializer(),
         "no_header_numpy": NoHeaderNumpySerializer(),
         "numpy": NumpySerializer(),
@@ -466,7 +529,7 @@ _SERIALIZERS = OrderedDict(
 )
 
 
-def _get_serializers(serializers: Optional[Dict[str, Serializer]]) -> Dict[str, Serializer]:
+def _get_serializers(serializers: Optional[dict[str, Serializer]]) -> dict[str, Serializer]:
     if serializers is None:
         serializers = {}
     serializers = OrderedDict(serializers)
