@@ -4,17 +4,39 @@ import math
 import os
 import shutil
 import tempfile
-from typing import Any, Dict, List, Optional, Tuple
+import time
+from typing import Any, Callable, Optional
 
 import numpy as np
 
-from litdata.constants import _DEFAULT_CACHE_DIR, _DEFAULT_LIGHTNING_CACHE_DIR, _INDEX_FILENAME
+from litdata.constants import _DEFAULT_CACHE_DIR, _DEFAULT_LIGHTNING_CACHE_DIR, _INDEX_FILENAME, _LITDATA_CACHE_DIR
 from litdata.streaming.downloader import get_downloader
 from litdata.streaming.item_loader import BaseItemLoader, TokensLoader
 from litdata.streaming.resolver import Dir, _resolve_dir
 from litdata.utilities.subsample import shuffle_lists_together, subsample_filenames_and_roi
 
 
+def wait_for_predicate(
+    predicate: Callable[[], bool],
+    timeout: float,
+) -> bool:
+    """Wait until the given predicate becomes True or the timeout expires.
+
+    Args:
+        predicate: A function returning a boolean condition to check.
+        timeout: Maximum time (in seconds) to wait.
+
+    Returns:
+        True if predicate became True within timeout, else False.
+    """
+    start = time.time()
+    while time.time() - start < timeout:
+        if predicate():
+            return True
+        time.sleep(0.01)
+    return False
+
+
 def subsample_streaming_dataset(
     input_dir: Dir,
     cache_dir: Optional[Dir] = None,
@@ -22,9 +44,11 @@ def subsample_streaming_dataset(
     subsample: float = 1.0,
     shuffle: bool = False,
     seed: int = 42,
-    storage_options: Optional[Dict] = {},
+    storage_options: Optional[dict] = {},
+    session_options: Optional[dict] = {},
     index_path: Optional[str] = None,
-) -> Tuple[List[str], List[Tuple[int, int]]]:
+    fnmatch_pattern: Optional[str] = None,
+) -> tuple[list[str], list[tuple[int, int]]]:
     """Subsample streaming dataset.
 
     But before doing that, we will do some preprocessing:
@@ -35,8 +59,8 @@ def subsample_streaming_dataset(
     - Once chunks are ready, subsample (chunk filenames, region_of_interest).
 
     """
-    subsampled_files: List[str] = []
-    roi: List[Tuple[int, int]] = []
+    subsampled_files: list[str] = []
+    roi: list[tuple[int, int]] = []
 
     # Make sure input_dir contains cache path and remote url
     if _should_replace_path(input_dir.path):
@@ -44,6 +68,7 @@ def subsample_streaming_dataset(
             input_dir=input_dir.path if input_dir.path else input_dir.url,
             cache_dir=cache_dir.path if cache_dir else None,
             storage_options=storage_options,
+            session_options=session_options,
             index_path=index_path,
         )
         if cache_path is not None:
@@ -59,10 +84,18 @@ def subsample_streaming_dataset(
         if index_path is not None:
             copy_index_to_cache_index_filepath(index_path, cache_index_filepath)
         else:
-            downloader = get_downloader(input_dir.url, input_dir.path, [], storage_options)
+            # Merge data_connection_id from resolved directory into storage_options for R2 connections
+            merged_storage_options = storage_options.copy() if storage_options is not None else {}
+            if hasattr(input_dir, "data_connection_id") and input_dir.data_connection_id:
+                merged_storage_options["data_connection_id"] = input_dir.data_connection_id
+
+            downloader = get_downloader(input_dir.url, input_dir.path, [], merged_storage_options, session_options)
             downloader.download_file(os.path.join(input_dir.url, _INDEX_FILENAME), cache_index_filepath)
 
-    if not os.path.exists(input_dir.path):
+    def path_exists(p: str) -> bool:
+        return wait_for_predicate(lambda: os.path.exists(p), timeout=0.5)
+
+    if not path_exists(input_dir.path):
         raise FileNotFoundError(f"The provided dataset path `{input_dir.path}` does not exist.")
 
     if os.path.exists(os.path.join(input_dir.path, _INDEX_FILENAME)):
@@ -75,6 +108,11 @@ def subsample_streaming_dataset(
             "\n HINT: Did you successfully optimize a dataset to the provided `input_dir`?"
         )
 
+    if fnmatch_pattern is not None:
+        from fnmatch import fnmatch
+
+        original_chunks = [chunk for chunk in original_chunks if fnmatch(chunk["filename"], fnmatch_pattern)]
+
     assert len(original_chunks) > 0, f"No chunks found in the `{input_dir}/index.json` file"
 
     # create a (chunk_start, chunk_end) list to indicate our subsample from where we can read.
@@ -85,8 +123,8 @@ def subsample_streaming_dataset(
 
         return subsampled_files, roi
 
-    final_files: List[str] = []
-    final_roi: List[Tuple[int, int]] = []
+    final_files: list[str] = []
+    final_roi: list[tuple[int, int]] = []
 
     random_seed_sampler = None
     if shuffle:
@@ -126,12 +164,14 @@ def _should_replace_path(path: Optional[str]) -> bool:
         or path.startswith("/teamspace/s3_folders/")
         or path.startswith("/teamspace/gcs_folders/")
         or path.startswith("/teamspace/gcs_connections/")
+        or path.startswith("/teamspace/lightning_storage/")
     )
 
 
 def _read_updated_at(
     input_dir: Optional[Dir],
-    storage_options: Optional[Dict] = {},
+    storage_options: Optional[dict] = {},
+    session_options: Optional[dict] = {},
     index_path: Optional[str] = None,
 ) -> str:
     """Read last updated timestamp from index.json file."""
@@ -139,9 +179,10 @@ def _read_updated_at(
     index_json_content = None
     assert isinstance(input_dir, Dir)
 
+    # Try to read index.json locally
     if input_dir.path is not None and os.path.exists(os.path.join(input_dir.path, _INDEX_FILENAME)):
-        # read index.json file and read last_updation_timestamp
         index_json_content = load_index_file(input_dir.path)
+    # Try to read index.json remotely
     elif input_dir.url is not None:
         assert input_dir.url is not None
         # download index.json file and read last_updation_timestamp
@@ -150,7 +191,7 @@ def _read_updated_at(
             if index_path is not None:
                 copy_index_to_cache_index_filepath(index_path, temp_index_filepath)
             else:
-                downloader = get_downloader(input_dir.url, tmp_directory, [], storage_options)
+                downloader = get_downloader(input_dir.url, tmp_directory, [], storage_options, session_options)
                 downloader.download_file(os.path.join(input_dir.url, _INDEX_FILENAME), temp_index_filepath)
             index_json_content = load_index_file(tmp_directory)
 
@@ -161,11 +202,14 @@ def _read_updated_at(
 
 
 def _clear_cache_dir_if_updated(input_dir_hash_filepath: str, updated_at_hash: str) -> None:
-    """Clear cache dir if it is updated.
+    """Clear the cache directory if it is outdated.
 
-    If last_updated has changed and /cache/chunks/{HASH(input_dir.url)} isn't empty, we remove all the files and then
-    create the cache.
+    If the directory at `input_dir_hash_filepath` exists and does not contain only a single subdirectory named
+    `updated_at_hash`, the entire directory is deleted to prevent using stale or partial cache data.
 
+    Args:
+        input_dir_hash_filepath (str): Path to the hashed cache directory (e.g., /cache/chunks/{HASH(input_dir.url)}).
+        updated_at_hash (str): The expected hash or timestamp for the current dataset state.
     """
     if os.path.exists(input_dir_hash_filepath):
         # check if it only contains one directory with updated_at_hash
@@ -174,35 +218,58 @@ def _clear_cache_dir_if_updated(input_dir_hash_filepath: str, updated_at_hash: s
             shutil.rmtree(input_dir_hash_filepath)
 
 
+def generate_md5_hash(value: str) -> str:
+    """Generate an MD5 hash for the given string value.
+
+    Args:
+        value (str): The input string to hash.
+
+    Returns:
+        str: The hexadecimal MD5 hash of the input string.
+    """
+    return hashlib.md5(value.encode()).hexdigest()  # noqa: S324
+
+
+def get_default_cache_dir() -> str:
+    """Return the default cache directory, using the Lightning cloud cache if running in a Lightning environment.
+
+    Returns:
+        str: The resolved default cache root directory.
+    """
+    if _LITDATA_CACHE_DIR is not None:
+        return _LITDATA_CACHE_DIR
+    is_lightning_cloud = "LIGHTNING_CLUSTER_ID" in os.environ and "LIGHTNING_CLOUD_PROJECT_ID" in os.environ
+    return _DEFAULT_LIGHTNING_CACHE_DIR if is_lightning_cloud else _DEFAULT_CACHE_DIR
+
+
 def _try_create_cache_dir(
     input_dir: Optional[str],
     cache_dir: Optional[str] = None,
-    storage_options: Optional[Dict] = {},
+    storage_options: Optional[dict] = {},
+    session_options: Optional[dict] = {},
     index_path: Optional[str] = None,
 ) -> Optional[str]:
+    """Prepare and return the cache directory for a dataset."""
     resolved_input_dir = _resolve_dir(input_dir)
-    updated_at = _read_updated_at(resolved_input_dir, storage_options, index_path)
+    updated_at = _read_updated_at(resolved_input_dir, storage_options, session_options, index_path)
 
+    # Fallback to a hash of the input_dir if updated_at is "0"
     if updated_at == "0" and input_dir is not None:
-        updated_at = hashlib.md5(input_dir.encode()).hexdigest()  # noqa: S324
+        updated_at = generate_md5_hash(input_dir)
 
-    dir_url_hash = hashlib.md5((resolved_input_dir.url or "").encode()).hexdigest()  # noqa: S324
+    dir_url_hash = generate_md5_hash(resolved_input_dir.url or "")
 
-    if "LIGHTNING_CLUSTER_ID" not in os.environ or "LIGHTNING_CLOUD_PROJECT_ID" not in os.environ:
-        input_dir_hash_filepath = os.path.join(cache_dir or _DEFAULT_CACHE_DIR, dir_url_hash)
-        _clear_cache_dir_if_updated(input_dir_hash_filepath, updated_at)
-        cache_dir = os.path.join(input_dir_hash_filepath, updated_at)
-        os.makedirs(cache_dir, exist_ok=True)
-        return cache_dir
+    # Determine the cache directory, preferring user-provided cache_dir if given
+    cache_dir = cache_dir if cache_dir is not None else get_default_cache_dir()
 
-    input_dir_hash_filepath = os.path.join(cache_dir or _DEFAULT_LIGHTNING_CACHE_DIR, dir_url_hash)
-    _clear_cache_dir_if_updated(input_dir_hash_filepath, updated_at)
-    cache_dir = os.path.join(input_dir_hash_filepath, updated_at)
+    input_dir_hash_path = os.path.join(cache_dir, dir_url_hash)
+    _clear_cache_dir_if_updated(input_dir_hash_path, updated_at)
+    cache_dir = os.path.join(input_dir_hash_path, updated_at)
     os.makedirs(cache_dir, exist_ok=True)
     return cache_dir
 
 
-def generate_roi(chunks: List[Dict[str, Any]], item_loader: Optional[BaseItemLoader] = None) -> List[Tuple[int, int]]:
+def generate_roi(chunks: list[dict[str, Any]], item_loader: Optional[BaseItemLoader] = None) -> list[tuple[int, int]]:
     """Generates default region_of_interest for chunks."""
     roi = []
 
@@ -217,7 +284,7 @@ def generate_roi(chunks: List[Dict[str, Any]], item_loader: Optional[BaseItemLoa
     return roi
 
 
-def load_index_file(input_dir: str) -> Dict[str, Any]:
+def load_index_file(input_dir: str) -> dict[str, Any]:
     """Load index file from the specified input directory.
 
     This function supports loading both chunk-based and mds shard-based index files.
@@ -247,7 +314,7 @@ def load_index_file(input_dir: str) -> Dict[str, Any]:
         raise FileNotFoundError(f"Index file not found at {index_filepath}.")
 
 
-def adapt_mds_shards_to_chunks(data: Dict[str, Any]) -> Dict[str, Any]:
+def adapt_mds_shards_to_chunks(data: dict[str, Any]) -> dict[str, Any]:
     """Adapt mds shard-based index data to chunk-based format for compatibility.
     For more details about MDS, refer to the MosaicML Streaming documentation: https://github.com/mosaicml/streaming.
 
@@ -296,7 +363,7 @@ def copy_index_to_cache_index_filepath(index_path: str, cache_index_filepath: st
     """Copy Index file from index_path to cache_index_filepath."""
     # If index_path is a directory, append "index.json"
     if os.path.isdir(index_path):
-        index_path = os.path.join(index_path, "index.json")
+        index_path = os.path.join(index_path, _INDEX_FILENAME)
     # Check if the file exists before copying
     if not os.path.isfile(index_path):
         raise FileNotFoundError(f"Index file not found: {index_path}")
